import pandas as pd
import numpy as np

from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis, calculate_time_constant_with_optimization, \
    calculate_time_constant_from_log
from lhcsmapi.pyedsl.QueryBuilder import flatten_nested_lists


def first(x):
    return x.iloc[0]


def first20mean(x):
    return np.mean(x.iloc[0:20])


def last(x):
    return x.iloc[-1]


def last20mean(x):
    return np.mean(x.iloc[-20:-1])


def tau_charge(df):
    exponent = 1
    slice_squared_df = df.pow(exponent)
    return ExpSignalAnalysis.calculate_characteristic_time_in_seconds_with_integral(slice_squared_df, exponent)


def tau_energy(df):
    exponent = 2
    slice_squared_df = df.pow(exponent)
    return ExpSignalAnalysis.calculate_characteristic_time_in_seconds_with_integral(slice_squared_df, exponent)


def tau_lin_reg(df):
    return calculate_time_constant_from_log(df)


def tau_exp_fit(df):
    return calculate_time_constant_with_optimization(df)


class Feature(object):
    def set_dfs(self, dfs):
        dfs = dfs if isinstance(dfs, (list, tuple)) else [dfs]
        dfs = flatten_nested_lists(dfs)
        # sort according to the column name
        self.dfs = dfs


class FeatureBuilder(object):

    def __init__(self):
        self.feature = Feature()

    def with_signal(self, *dfs):
        self.feature.set_dfs(dfs)

        return FeatureBuilderSignal(self.feature)


class FeatureBuilderSignal(object):

    __name_to_function = {"first": first, "first20mean": first20mean, "last": last, "last20mean": last20mean,
                          "max": max, "min": min, "median": np.median, "std": np.std, "mean": np.mean,
                          "tau_charge": tau_charge, "tau_energy": tau_energy, "tau_lin_reg": tau_lin_reg,
                          "tau_exp_fit": tau_exp_fit}

    def __init__(self, feature):
        self.feature = feature

    def calculate_features(self, features, index=0):
        functions_to_apply = [self.__name_to_function[x] for x in features]

        features_dfs = []
        for df in self.feature.dfs:
            features_df = df.aggregate(functions_to_apply).T
            feature_col_names = {col: '{}:{}'.format(features_df.index[0], col) for col in features_df.columns}
            features_df = features_df.rename(columns=feature_col_names)
            features_df.index = [index]
            features_dfs.append(features_df)

        return pd.concat(features_dfs, axis=1)
